/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is ThumbsDown.
 *
 * The Initial Developer of the Original Code is
 * Atte Kemppilä.
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 * 
 * ***** END LICENSE BLOCK ***** */


const CLASS_ID = Components.ID("5cb8d346-b628-11dc-8314-0800200c9a66");
const CLASS_NAME = "ThumbsDown Script Runner";
const CONTRACT_ID = "@thumbsdown.mozdev.org/script-runner;1";
const CATEGORY_ENTRY = "thumbsdown";

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;

const MAX_INT = 4294967295;  // 2^32 - 1

// subject: scriptRunner; data: old scripts count
const TOPIC_NEW     = "thumbsdown-script-new";
// subject: scriptRunner; data: list of removed indexes in *descending* order (as array.toSource())
const TOPIC_DESTROY = "thumbsdown-script-destroy";
// subject: script; data: null
const TOPIC_CHANGE  = "thumbsdown-script-change";

const SCRIPT_ERROR = "thumbsdown_ScriptError";

const STATE_QUEUED     = 0;
const STATE_EVALUATING = 1;
const STATE_LOADING    = 2;
const STATE_CALLBACK   = 3;
const STATE_CANCELLED  = 4;
const STATE_SUCCESS    = 5;
const STATE_ERROR      = 6;

const TYPE_DIRECT = 0;
const TYPE_REGEXP = 1;
const TYPE_CUSTOM = 2;


var gScriptRunner = null;
var gScriptUtils = {};


function ScriptRunner()
{
    this.wrappedJSObject = this;
    this.scripts = [];
    this.scriptIndex = null;
    this.count = {running: 0, cancelled: 0, success: 0, error: 0};
    
    var loader = Cc["@mozilla.org/moz/jssubscript-loader;1"].getService(Ci.mozIJSSubScriptLoader);
    loader.loadSubScript("chrome://global/content/nsUserSettings.js");
    loader.loadSubScript("chrome://thumbsdown/content/resourceloader.js");
    loader.loadSubScript("chrome://thumbsdown/content/script_utils.js", gScriptUtils);
    loader.loadSubScript("chrome://thumbsdown/content/util.js");
    
    this.scriptMgr = Cc["@thumbsdown.mozdev.org/script-manager;1"].getService(Ci.nsISupports).wrappedJSObject;
    
    var obsSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    obsSvc.addObserver(this, TOPIC_CHANGE, true);
}

ScriptRunner.prototype =
{
    // nsISupports
    QueryInterface: function(iid)
    {
        if (iid.equals(Ci.nsIObserver) ||
            iid.equals(Ci.nsISupports) ||
            iid.equals(Ci.nsISupportsWeakReference))
            return this;
        throw Cr.NS_ERROR_NO_INTERFACE;
    },
    
    // nsIObserver
    observe: function(subject, topic, data)
    {
        if (topic != TOPIC_CHANGE)
            throw new Error("unknown topic: " + topic);
        var script = subject.wrappedJSObject;
        
        if ((script.oldState == STATE_QUEUED) && script.isRunning)
            this.count.running++;
        else if (script.isFinished)
        {
            this.count.running--;
            switch (script.state)
            {
                case STATE_CANCELLED: this.count.cancelled++; break;
                case STATE_SUCCESS:   this.count.success++; break;
                case STATE_ERROR:     this.count.error++; break;
            }
            this._startNextScript();
        }
    },
    
    get length() { return this.scripts.length; },
    
    addScripts: function(scriptArgs, forceStart)
    {
        this._initScriptIndex();
        var oldCount = this.scripts.length;
        var newScripts = [];
        for (var i = 0; i < scriptArgs.length; i++)
        {
            var args = scriptArgs[i];
            var script = new Script(args.name, args.code, args.documentUrl, args.thumbnailUrl, args.linkUrl, args.callback);
            newScripts.push(script);
            this.scripts.push(script);
            this.scriptIndex[script.id] = oldCount + i;
        }
        if (scriptArgs.length == 0)
            return;
        var obsSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        obsSvc.notifyObservers(this, TOPIC_NEW, oldCount);
        if (forceStart)
            for each (script in newScripts)
                schedule(script, script.start);
        else
            this._startNextScript();
    },
    
    removeScripts: function(indexes)
    {
        var removed = [];
        indexes = indexes.concat();
        indexes.reverse();
        for each (index in indexes)
        {
            if (! ((0 <= index) && (index < this.scripts.length) && (! this.scripts[index].isRunning)))
                continue;
            var script = this.scripts.splice(index, 1)[0];
            switch (script.state)
            {
                case STATE_CANCELLED: this.count.cancelled--; break;
                case STATE_SUCCESS:   this.count.success--; break;
                case STATE_ERROR:     this.count.error--; break;
            }
            removed.push(index);
        }
        if (removed.length == 0)
            return;
        this.scriptIndex = null;
        var obsSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
        obsSvc.notifyObservers(this, TOPIC_DESTROY, removed.toSource());
    },
    
    openWindow: function(focus)
    {
        var uri = "chrome://thumbsdown/content/scriptrunner.xul";
        var name = "thumbsdown_scriptRunner";
        var features = "chrome";
        var windowsMediator = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);
        var win = windowsMediator.getMostRecentWindow(name);
        if (win)
        {
            if (focus)
                win.focus();
        }
        else
        {
            var windowWatcher = Cc["@mozilla.org/embedcomp/window-watcher;1"].getService(Ci.nsIWindowWatcher);
            var parent = null;
            var args = null;
            windowWatcher.openWindow(parent, uri, name, features, args);
        }
    },
    
    _startNextScript: function()
    {
        if (this.count.running > 0)
            return;
        for each (script in this.scripts)
            if (script.isQueued)
            {
                schedule(script, script.start);
                return;
            }
    },
    
    _getScriptIndex: function(script)
    {
        this._initScriptIndex();
        var index = this.scriptIndex[script.id];
        return index;
    },
    
    _initScriptIndex: function()
    {
        if (this.scriptIndex)
            return;
        this.scriptIndex = {};
        for (var i = 0; i < this.scripts.length; i++)
            this.scriptIndex[this.scripts[i].id] = i;
    }
}


function Script(name, code, documentUrl, thumbnailUrl, linkUrl, callback)
{
    this.wrappedJSObject = this;
    this.id = Script.nextId();
    this.name = name;
    this.code = code;
    this.documentUrl = documentUrl;
    this.thumbnailUrl = thumbnailUrl;
    this.linkUrl = linkUrl;
    this.callback = callback;
    this._state = STATE_QUEUED;
    this.oldState = null;
    this.startTime = null;
    this.endTime = null;
    this.scriptState = 0;
    this.resourceLoader = null;
    this.result = null;
    this.referer = null;
    this.error = null;
}

Script._nextId = 1;
Script.nextId = function()
{
    return String((Script._nextId <= 0) ? 1 : Script._nextId++);
};

Script.prototype =
{
    get index() { return gScriptRunner._getScriptIndex(this); },
    
    get state() { return this._state; },
    set state(value)
    {
        if (value != this._state)
        {
            this.oldState = this._state;
            this._state = value;
            var obsSvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
            obsSvc.notifyObservers(this, TOPIC_CHANGE, null);
        }
    },
    
    get isQueued()    { return (this._state == STATE_QUEUED); },
    get isCancelled() { return (this._state == STATE_CANCELLED); },
    get isSuccess()   { return (this._state == STATE_SUCCESS); },
    get isError()     { return (this._state == STATE_ERROR); },
    get isRunning()   { return ((this._state == STATE_EVALUATING) || (this._state == STATE_LOADING) || (this._state == STATE_CALLBACK)); },
    get isFinished()  { return ((this._state == STATE_CANCELLED)  || (this._state == STATE_SUCCESS) || (this._state == STATE_ERROR)); },
    
    start: function()
    {
        if (! this.isQueued)
            return;
        this.startTime = new Date();
        this.endTime = null;
        this.result = null;
        this.error = null;
        this._eval(0, this.documentUrl);
    },
    
    cancel: function()
    {
        if (! this.isRunning)
            return;
        if (this.resourceLoader)
        {
            this.resourceLoader.cancel();
            this.resourceLoader = null;
        }
        this.endTime = new Date();
        this.state = STATE_CANCELLED;
    },
    
    _eval: function(scriptState, docUrl, doc)
    {
        try
        {
            if (this.isCancelled)
                return;
            this.scriptState = scriptState;
            this.state = STATE_EVALUATING;
            var sandbox = this._initSandbox(docUrl, doc);
            var result = this._evalInSandbox(this.code, sandbox);
            this._processResult(docUrl, result);
        }
        catch (e)
        {
            // script errors are already reported in _evalInSandbox()
            if (e.name != SCRIPT_ERROR)
                Cu.reportError(e);
            this._handleError(e);
        }
    },
    
    _initSandbox: function(docUrl, doc)
    {
        var sandbox = new Cu.Sandbox(docUrl);
        sandbox.state = this.scriptState;
        sandbox.thumbnailUrl = this.thumbnailUrl;
        sandbox.linkUrl = this.linkUrl;
        if (doc)
            sandbox.document = doc;
        for (let [name, value] in Iterator(gScriptUtils))
            sandbox[name] = value;
        return sandbox;
    },
    
    _evalInSandbox: function(code, sandbox)
    {
        try
        {
            // lineFinder trick is taken from GreaseMonkey
            // workaround for https://bugzilla.mozilla.org/show_bug.cgi?id=307984
            var lineFinder = new Error();
            Cu.evalInSandbox(code, sandbox);
            throw new Error("abnormal exit");
        }
        catch (e if (typeof(e) == "string"))
        {
            return e;
        }
        catch (e)
        {
            // Line number is reported as max int in edge cases.
            // Sometimes the right one is in the "location", instead. Look there.
            var line = e.lineNumber;
            if (line == MAX_INT)
                if (e.location && e.location.lineNumber)
                    line = e.location.lineNumber;
                else
                    line = 0;
            line = line - lineFinder.lineNumber - 1;
            
            // one cannot modify error and report it with Cu.reportError (it shows original values)
            // therefore we report it manually in error console
            var consoleSvc = Cc["@mozilla.org/consoleservice;1"].getService(Ci.nsIConsoleService);
            var scriptError = Cc["@mozilla.org/scripterror;1"].createInstance(Ci.nsIScriptError);
            var msg = SCRIPT_ERROR + ": " + e.name + ": " + e.message;
            var srcName = "ThumbsDown Script: " + this.name;
            var srcLine = null;
            var column = e.columnNumber;
            var flag = Ci.nsIScriptError.errorFlag;
            var category = null;
            scriptError.init(msg, srcName, srcLine, line, column, flag, category);
            consoleSvc.logMessage(scriptError);
            
            // error still needs to be thrown, just not reported
            e.message = e.name + ": " + e.message + " (line: " + line + ")";
            e.name = SCRIPT_ERROR;
            e.fileName = srcName;
            e.lineNumber = line;
            throw e;
        }
    },
    
    _processResult: function(docUrl, result)
    {
        var index = result.indexOf(":");
        if (index == -1)
            throw new Error("invalid script result format: '" + result + "'");
        var op = result.substr(0, index);
        var arg = result.substr(index + 1);
        switch (op)
        {
            case "LOAD":
                this._load(arg, docUrl);
                break;
            case "REDIRECT":
                this._redirect(arg, docUrl);
                break;
            case "RETURN":
                this._doCallback(arg, docUrl);
                break;
            default:
                throw new Error("invalid script result operation: '" + op + "'");
        }
    },
    
    _load: function(url, referer)
    {
        var ioSvc = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
        var uri = ioSvc.newURI(url, null, null);
        var headers = {"Referer": referer};
        var onSuccess = getCallback(this, this._onLoadSuccess);
        var onError = getCallback(this, this._onLoadError);
        var timeout = nsPreferences.getIntPref("extensions.thumbsdown.scriptrunner.requestTimeout");
        this.resourceLoader = new ResourceLoader("GET", uri, headers, onSuccess, onError, timeout);
        this.resourceLoader.load();
        this.state = STATE_LOADING;
    },
    
    _onLoadSuccess: function()
    {
        if (this.isCancelled)
            return;
        var docUrl = this.resourceLoader.uri.spec;
        var doc = this.resourceLoader.content;
        this.resourceLoader.destroy();
        this.resourceLoader = null;
        this._eval(this.scriptState + 1, docUrl, doc);
    },
    
    _onLoadError: function()
    {
        if (this.isCancelled)
            return;
        // TODO: error message from resource -> finish(new Error(msg))
        var msg;
        if (this.resourceLoader.isTimeout)
            msg = "request timeout";
        else
            msg = "load error";
        this.resourceLoader.destroy();
        this.resourceLoader = null;
        this._handleError(new Error(msg));
    },
    
    _redirect: function(url, docUrl)
    {
        var script = gScriptRunner.scriptMgr.lookup(url, null);
        if (! script)
            throw new Error("failed redirect: " + url);
        switch (script.type)
        {
            case TYPE_DIRECT:
                this._doCallback(url, docUrl);
                break;
            case TYPE_REGEXP:
                this._doCallback(url.replace(script.pattern, script.template), url);
                break;
            case TYPE_CUSTOM:
                var args = {};
                args.name = script.name;
                args.code = script.script;
                args.documentUrl = docUrl;
                args.linkUrl = url;
                args.callback = this.callback;
                gScriptRunner.addScripts([args], true);
                this.child = gScriptRunner.scripts[gScriptRunner.scripts.length - 1];
                this.child.parent = this;
                schedule(this, this._handleSuccess);
                break;
            default:
                throw new Error("unreachable code");
        }
    },
    
    _doCallback: function(url, referer)
    {
        this.result = url;
        this.referer = referer;
        this.state = STATE_CALLBACK;
        this.callback(this);
        this._handleSuccess();
    },
    
    _handleSuccess: function()
    {
        this.endTime = new Date();
        this.state = STATE_SUCCESS;
    },
    
    _handleError: function(error)
    {
        this.result = null;
        this.error = error;
        this.endTime = new Date();
        this.state = STATE_ERROR;
    }
}


// nsIFactory
var ScriptRunnerFactory =
{
    createInstance: function(outer, iid)
    {
        if (outer != null)
            throw Cr.NS_ERROR_NO_AGGREGATION;
        if (gScriptRunner == null)
            gScriptRunner = new ScriptRunner();
        return gScriptRunner.QueryInterface(iid);
    }
};

// nsIModule
var ScriptRunnerModule =
{
    registerSelf: function(compMgr, fileSpec, location, type)
    {
        // register
        compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
        compMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, fileSpec, location, type);
        // add start-up observer
        //var catMgr = Cc["@mozilla.org/categorymanager;1"].getService(Ci.nsICategoryManager);
        //catMgr.addCategoryEntry("xpcom-startup", CATEGORY_ENTRY, CONTRACT_ID, true, true);
    },
    
    unregisterSelf: function(compMgr, fileSpec, location)
    {
        // remove start-up observer
        //var catMgr = Cc["@mozilla.org/categorymanager;1"].getService(Ci.nsICategoryManager);
        //catMgr.deleteCategoryEntry("xpcom-startup", CATEGORY_ENTRY, true);
        // unregister
        compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
        compMgr.unregisterFactoryLocation(CLASS_ID, location);
    },
    
    getClassObject: function(compMgr, cid, iid)
    {
        if (! cid.equals(CLASS_ID))
            throw Cr.NS_ERROR_NO_INTERFACE;
        if (! iid.equals(Ci.nsIFactory))
            throw Cr.NS_ERROR_NOT_IMPLEMENTED;
        return ScriptRunnerFactory;
    },
    
    canUnload: function(compMgr)
    {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec)
{
    return ScriptRunnerModule;
}
